@@ -73,6 +73,7 @@ modules = [
     "girlswithmuscle",
     "gofile",
     "hatenablog",
+    "hdoujin",
     "hentai2read",
     "hentaicosplays",
     "hentaifoundry",
@@ -88,6 +89,7 @@ modules = [
     "imagefap",
     "imgbb",
     "imgbox",
+    "imgpile",
     "imgth",
     "imgur",
     "imhentai",
@@ -113,11 +115,14 @@ modules = [
     "lynxchan",
     "madokami",
     "mangadex",
+    "mangafire",
     "mangafox",
     "mangahere",
     "manganelo",
     "mangapark",
     "mangaread",
+    "mangareader",
+    "mangataro",
     "mangoxo",
     "misskey",
     "motherless",
@@ -163,6 +168,7 @@ modules = [
     "rule34us",
     "rule34vault",
     "rule34xyz",
+    "s3ndpics",
     "saint",
     "sankaku",
     "sankakucomplex",
@@ -188,6 +194,7 @@ modules = [
     "tcbscans",
     "telegraph",
     "tenor",
+    "thehentaiworld",
     "tiktok",
     "tmohentai",
     "toyhouse",
