@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
-# Copyright 2019-2023 Mike Fährmann
+# Copyright 2019-2025 Mike Fährmann
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 2 as
@@ -10,7 +10,7 @@
 import os
 import sys
 import unittest
-from unittest.mock import Mock, mock_open, patch
+from unittest.mock import Mock, mock_open, patch, call
 
 import shutil
 import logging
@@ -20,8 +20,8 @@ import collections
 from datetime import datetime
 
 sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
-from gallery_dl import extractor, output, path  # noqa E402
-from gallery_dl import postprocessor, config  # noqa E402
+from gallery_dl import extractor, output, path, util, exception  # noqa E402
+from gallery_dl import postprocessor, config, archive  # noqa E402
 from gallery_dl.postprocessor.common import PostProcessor  # noqa E402
 
 
@@ -39,7 +39,7 @@ class FakeJob():
         self.get_logger = logging.getLogger
         self.hooks = collections.defaultdict(list)
 
-    def register_hooks(self, hooks, options):
+    def register_hooks(self, hooks, options=None):
         for hook, callback in hooks.items():
             self.hooks[hook].append(callback)
 
@@ -52,7 +52,7 @@ class TestPostprocessorModule(unittest.TestCase):
     def test_find(self):
         for name in (postprocessor.modules):
             cls = postprocessor.find(name)
-            self.assertEqual(cls.__name__, name.capitalize() + "PP")
+            self.assertEqual(cls.__name__, f"{name.capitalize()}PP")
             self.assertIs(cls.__base__, PostProcessor)
 
         self.assertEqual(postprocessor.find("foo"), None)
@@ -129,15 +129,15 @@ class ClassifyTest(BasePostprocessorTest):
         self._trigger(("prepare",))
         self.pathfmt.build_path()
         path = os.path.join(self.dir.name, "test", "Pictures")
-        self.assertEqual(self.pathfmt.path, path + "/file.jpg")
-        self.assertEqual(self.pathfmt.realpath, path + "/file.jpg")
+        self.assertEqual(self.pathfmt.path, f"{path}/file.jpg")
+        self.assertEqual(self.pathfmt.realpath, f"{path}/file.jpg")
 
         self.pathfmt.set_extension("mp4")
         self._trigger(("prepare",))
         self.pathfmt.build_path()
         path = os.path.join(self.dir.name, "test", "Video")
-        self.assertEqual(self.pathfmt.path, path + "/file.mp4")
-        self.assertEqual(self.pathfmt.realpath, path + "/file.mp4")
+        self.assertEqual(self.pathfmt.path, f"{path}/file.mp4")
+        self.assertEqual(self.pathfmt.realpath, f"{path}/file.mp4")
 
     def test_classify_noop(self):
         pp = self._create()
@@ -169,8 +169,26 @@ class ClassifyTest(BasePostprocessorTest):
         self._trigger(("prepare",))
         self.pathfmt.build_path()
         path = os.path.join(self.dir.name, "test", "foo", "bar")
-        self.assertEqual(self.pathfmt.path, path + "/file.foo")
-        self.assertEqual(self.pathfmt.realpath, path + "/file.foo")
+        self.assertEqual(self.pathfmt.path, f"{path}/file.foo")
+        self.assertEqual(self.pathfmt.realpath, f"{path}/file.foo")
+
+
+class DirectoryTest(BasePostprocessorTest):
+
+    def test_default(self):
+        self._create()
+
+        path = os.path.join(self.dir.name, "test")
+        self.assertEqual(self.pathfmt.realdirectory, f"{path}/")
+        self.assertEqual(self.pathfmt.realpath, f"{path}/file.ext")
+
+        self.pathfmt.kwdict["category"] = "custom"
+        self._trigger()
+
+        path = os.path.join(self.dir.name, "custom")
+        self.assertEqual(self.pathfmt.realdirectory, f"{path}/")
+        self.pathfmt.build_path()
+        self.assertEqual(self.pathfmt.realpath, f"{path}/file.ext")
 
 
 class ExecTest(BasePostprocessorTest):
@@ -187,11 +205,16 @@ class ExecTest(BasePostprocessorTest):
             self._trigger(("after",))
 
         p.assert_called_once_with(
-            "echo {0} {0} {1} {2} && rm {0};".format(
-                self.pathfmt.realpath,
-                self.pathfmt.realdirectory,
-                self.pathfmt.filename),
-            shell=True)
+            (f"echo "
+             f"{self.pathfmt.realpath} "
+             f"{self.pathfmt.realpath} "
+             f"{self.pathfmt.realdirectory} "
+             f"{self.pathfmt.filename} "
+             f"&& rm {self.pathfmt.realpath};"),
+            shell=True,
+            creationflags=0,
+            start_new_session=False,
+        )
         i.wait.assert_called_once_with()
 
     def test_command_list(self):
@@ -213,8 +236,48 @@ class ExecTest(BasePostprocessorTest):
                 self.pathfmt.realdirectory.upper(),
             ],
             shell=False,
+            creationflags=0,
+            start_new_session=False,
         )
 
+    def test_command_many(self):
+        self._create({
+            "commands": [
+                "echo {} {_path} {_directory} {_filename} && rm {};",
+                ["~/script.sh", "{category}", "\fE _directory.upper()"],
+            ]
+        })
+
+        with patch("gallery_dl.util.Popen") as p:
+            i = Mock()
+            i.wait.return_value = 0
+            p.return_value = i
+            self._trigger(("after",))
+
+        self.assertEqual(p.call_args_list, [
+            call(
+                (f"echo "
+                 f"{self.pathfmt.realpath} "
+                 f"{self.pathfmt.realpath} "
+                 f"{self.pathfmt.realdirectory} "
+                 f"{self.pathfmt.filename} "
+                 f"&& rm {self.pathfmt.realpath};"),
+                shell=True,
+                creationflags=0,
+                start_new_session=False,
+            ),
+            call(
+                [
+                    os.path.expanduser("~/script.sh"),
+                    self.pathfmt.kwdict["category"],
+                    self.pathfmt.realdirectory.upper(),
+                ],
+                shell=False,
+                creationflags=0,
+                start_new_session=False,
+            ),
+        ])
+
     def test_command_returncode(self):
         self._create({
             "command": "echo {}",
@@ -228,8 +291,9 @@ class ExecTest(BasePostprocessorTest):
             with self.assertLogs() as log:
                 self._trigger(("after",))
 
-        msg = ("WARNING:postprocessor.exec:'echo {}' returned with "
-               "non-zero exit status (123)".format(self.pathfmt.realpath))
+        msg = (f"WARNING:postprocessor.exec:"
+               f"'echo {self.pathfmt.realpath}' "
+               f"returned with non-zero exit status (123)")
         self.assertEqual(log.output[0], msg)
 
     def test_async(self):
@@ -246,6 +310,66 @@ class ExecTest(BasePostprocessorTest):
         self.assertTrue(p.called)
         self.assertFalse(i.wait.called)
 
+    @unittest.skipIf(util.WINDOWS, "not POSIX")
+    def test_session_posix(self):
+        self._create({
+            "session": True,
+            "command": ["echo", "foobar"],
+        })
+
+        with patch("gallery_dl.util.Popen") as p:
+            i = Mock()
+            i.wait.return_value = 0
+            p.return_value = i
+            self._trigger(("after",))
+
+        p.assert_called_once_with(
+            ["echo", "foobar"],
+            shell=False,
+            creationflags=0,
+            start_new_session=True,
+        )
+        i.wait.assert_called_once_with()
+
+    @unittest.skipIf(not util.WINDOWS, "not Windows")
+    def test_session_windows(self):
+        self._create({
+            "session": True,
+            "command": ["echo", "foobar"],
+        })
+
+        with patch("gallery_dl.util.Popen") as p:
+            i = Mock()
+            i.wait.return_value = 0
+            p.return_value = i
+            self._trigger(("after",))
+
+        import subprocess
+        p.assert_called_once_with(
+            ["echo", "foobar"],
+            shell=False,
+            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
+            start_new_session=False,
+        )
+        i.wait.assert_called_once_with()
+
+    def test_archive(self):
+        pp = self._create({
+            "command": ["echo", "foobar"],
+            "archive": ":memory:",
+            "event"  : "finalize",
+        })
+
+        self.assertIsInstance(pp.archive, archive.DownloadArchive)
+
+        with patch.object(pp.archive, "add") as m_aa, \
+                patch.object(pp.archive, "close") as m_ac:
+            self._trigger(("finalize",))
+        pp.archive.close()
+
+        m_aa.assert_called_once_with(self.pathfmt.kwdict)
+        m_ac.assert_called_once()
+
 
 class HashTest(BasePostprocessorTest):
 
@@ -324,12 +448,10 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realpath + ".JSON"
+        path = f"{self.pathfmt.realpath}.JSON"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
-        if sys.hexversion >= 0x3060000:
-            # python 3.4 & 3.5 have random order without 'sort: True'
-            self.assertEqual(self._output(m), """{
+        self.assertEqual(self._output(m), """{
     "category": "test",
     "filename": "file",
     "extension": "ext",
@@ -360,7 +482,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realpath + ".JSON"
+        path = f"{self.pathfmt.realpath}.JSON"
         m.assert_called_once_with(path, "a", encoding="UTF-8")
         self.assertEqual(self._output(m), """{\
 "_private" : "foo \\u30d0\\u30fc",\
@@ -381,7 +503,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realpath + ".txt"
+        path = f"{self.pathfmt.realpath}.txt"
         m.assert_called_once_with(path, "w", encoding="utf-8")
         self.assertEqual(self._output(m), "foo\nbar\nbaz\n")
 
@@ -450,6 +572,17 @@ class MetadataTest(BasePostprocessorTest):
         test({"mode": "custom", "format": "{foo}\n{missing}\n"})
         test({"format": "{foo}\n{missing}\n"})
 
+    def test_metadata_mode_print(self):
+        self._create(
+            {"mode": "print", "format": "{foo}\n{missing}"},
+            {"foo": "bar"},
+        )
+
+        with patch("sys.stdout", Mock()) as m:
+            self._trigger()
+
+        self.assertEqual(self._output(m), "bar\nNone\n")
+
     def test_metadata_extfmt(self):
         pp = self._create({
             "extension"       : "ignored",
@@ -461,7 +594,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "file.json"
+        path = f"{self.pathfmt.realdirectory}file.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_extfmt_2(self):
@@ -473,7 +606,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "file.2.EXT-data:tESt"
+        path = f"{self.pathfmt.realdirectory}file.2.EXT-data:tESt"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_directory(self):
@@ -484,7 +617,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "metadata/file.ext.json"
+        path = f"{self.pathfmt.realdirectory}metadata/file.ext.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_directory_2(self):
@@ -496,7 +629,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "metadata/file.json"
+        path = f"{self.pathfmt.realdirectory}metadata/file.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_directory_format(self):
@@ -508,7 +641,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "../json/12500/file.ext.json"
+        path = f"{self.pathfmt.realdirectory}../json/12500/file.ext.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_directory_empty(self):
@@ -519,7 +652,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "./file.ext.json"
+        path = f"{self.pathfmt.realdirectory}./file.ext.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_basedirectory(self):
@@ -528,7 +661,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.basedirectory + "file.ext.json"
+        path = f"{self.pathfmt.basedirectory}file.ext.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_basedirectory_custom(self):
@@ -552,7 +685,7 @@ class MetadataTest(BasePostprocessorTest):
         with patch("builtins.open", mock_open()) as m:
             self._trigger()
 
-        path = self.pathfmt.realdirectory + "test_file__meta_.data"
+        path = f"{self.pathfmt.realdirectory}test_file__meta_.data"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_meta_path(self):
@@ -563,7 +696,7 @@ class MetadataTest(BasePostprocessorTest):
         self._trigger()
 
         self.assertEqual(self.pathfmt.kwdict["_meta_path"],
-                         self.pathfmt.realpath + ".json")
+                         f"{self.pathfmt.realpath}.json")
 
     def test_metadata_stdout(self):
         self._create({"filename": "-", "indent": None, "sort": True})
@@ -652,7 +785,7 @@ class MetadataTest(BasePostprocessorTest):
         self.assertTrue(m.called)
         self.assertGreater(len(self._output(m)), 0)
 
-        path = self.pathfmt.realdirectory + "file.ext.json"
+        path = f"{self.pathfmt.realdirectory}file.ext.json"
         m.assert_called_once_with(path, "w", encoding="utf-8")
 
     def test_metadata_option_skip_false(self):
@@ -695,8 +828,23 @@ class MetadataTest(BasePostprocessorTest):
 }
 """)
 
-    @staticmethod
-    def _output(mock):
+    def test_archive(self):
+        pp = self._create({
+            "archive": ":memory:",
+            "event"  : "finalize",
+        })
+
+        self.assertIsInstance(pp.archive, archive.DownloadArchive)
+
+        with patch.object(pp.archive, "add") as m_aa, \
+                patch.object(pp.archive, "close") as m_ac:
+            self._trigger(("finalize",))
+        pp.archive.close()
+
+        m_aa.assert_called_once_with(self.pathfmt.kwdict)
+        m_ac.assert_called_once()
+
+    def _output(self, mock):
         return "".join(
             call[1][0]
             for call in mock.mock_calls
@@ -709,32 +857,32 @@ class MtimeTest(BasePostprocessorTest):
     def test_mtime_datetime(self):
         self._create(None, {"date": datetime(1980, 1, 1)})
         self._trigger()
-        self.assertEqual(self.pathfmt.kwdict["_mtime"], 315532800)
+        self.assertEqual(self.pathfmt.kwdict["_mtime_meta"], 315532800)
 
     def test_mtime_timestamp(self):
         self._create(None, {"date": 315532800})
         self._trigger()
-        self.assertEqual(self.pathfmt.kwdict["_mtime"], 315532800)
+        self.assertEqual(self.pathfmt.kwdict["_mtime_meta"], 315532800)
 
     def test_mtime_none(self):
         self._create(None, {"date": None})
         self._trigger()
-        self.assertNotIn("_mtime", self.pathfmt.kwdict)
+        self.assertNotIn("_mtime_meta", self.pathfmt.kwdict)
 
     def test_mtime_undefined(self):
         self._create(None, {})
         self._trigger()
-        self.assertNotIn("_mtime", self.pathfmt.kwdict)
+        self.assertNotIn("_mtime_meta", self.pathfmt.kwdict)
 
     def test_mtime_key(self):
         self._create({"key": "foo"}, {"foo": 315532800})
         self._trigger()
-        self.assertEqual(self.pathfmt.kwdict["_mtime"], 315532800)
+        self.assertEqual(self.pathfmt.kwdict["_mtime_meta"], 315532800)
 
     def test_mtime_value(self):
         self._create({"value": "{foo}"}, {"foo": 315532800})
         self._trigger()
-        self.assertEqual(self.pathfmt.kwdict["_mtime"], 315532800)
+        self.assertEqual(self.pathfmt.kwdict["_mtime_meta"], 315532800)
 
 
 class PythonTest(BasePostprocessorTest):
@@ -757,12 +905,41 @@ class PythonTest(BasePostprocessorTest):
         path = os.path.join(self.dir.name, "module.py")
         self._write_module(path)
 
-        self._create({"function": path + ":calc"}, {"_value": 12})
+        self._create({"function": f"{path}:calc"}, {"_value": 12})
 
         self.assertNotIn("_result", self.pathfmt.kwdict)
         self._trigger()
         self.assertEqual(self.pathfmt.kwdict["_result"], 24)
 
+    def test_eval(self):
+        self._create({"mode": "eval", "expression": "abort()"})
+
+        with self.assertRaises(exception.StopExtraction):
+            self._trigger()
+
+    def test_eval_auto(self):
+        self._create({"expression": "abort()"})
+
+        with self.assertRaises(exception.StopExtraction):
+            self._trigger()
+
+    def test_archive(self):
+        pp = self._create({
+            "expression": "True",
+            "archive"   : ":memory:",
+            "event"     : "finalize",
+        })
+
+        self.assertIsInstance(pp.archive, archive.DownloadArchive)
+
+        with patch.object(pp.archive, "add") as m_aa, \
+                patch.object(pp.archive, "close") as m_ac:
+            self._trigger(("finalize",))
+        pp.archive.close()
+
+        m_aa.assert_called_once_with(self.pathfmt.kwdict)
+        m_ac.assert_called_once()
+
     def _write_module(self, path):
         with open(path, "w") as fp:
             fp.write("""
@@ -814,7 +991,7 @@ class RenameTest(BasePostprocessorTest):
     def test_rename_skip(self):
         self._create({"from": "{id}.{extension}"}, {"id": 12345})
         path = self._prepare("12345.ext")
-        with open(path + "file.ext", "w"):
+        with open(f"{path}file.ext", "w"):
             pass
 
         with self.assertLogs("postprocessor.rename", level="WARNING") as cm:
@@ -833,7 +1010,7 @@ class ZipTest(BasePostprocessorTest):
         self.assertEqual(pp.path, self.pathfmt.realdirectory[:-1])
         self.assertEqual(pp.delete, True)
         self.assertEqual(pp.args, (
-            pp.path + ".zip", "a", zipfile.ZIP_STORED, True,
+            f"{pp.path}.zip", "a", zipfile.ZIP_STORED, True,
         ))
         self.assertTrue(pp.args[0].endswith("/test.zip"))
 
@@ -843,7 +1020,7 @@ class ZipTest(BasePostprocessorTest):
         self.assertEqual(pp.path, self.pathfmt.realdirectory[:-1])
         self.assertEqual(pp.delete, True)
         self.assertEqual(pp.args, (
-            pp.path + ".zip", "a", zipfile.ZIP_STORED, True,
+            f"{pp.path}.zip", "a", zipfile.ZIP_STORED, True,
         ))
         self.assertTrue(pp.args[0].endswith("/test.zip"))
 
@@ -855,7 +1032,7 @@ class ZipTest(BasePostprocessorTest):
         })
         self.assertEqual(pp.delete, False)
         self.assertEqual(pp.args, (
-            pp.path + ".cbz", "a", zipfile.ZIP_DEFLATED, True,
+            f"{pp.path}.cbz", "a", zipfile.ZIP_DEFLATED, True,
         ))
         self.assertTrue(pp.args[0].endswith("/test.cbz"))
 
@@ -869,7 +1046,7 @@ class ZipTest(BasePostprocessorTest):
 
             # write dummy file with 3 different names
             for i in range(3):
-                name = "file{}.ext".format(i)
+                name = f"file{i}.ext"
                 self.pathfmt.temppath = file.name
                 self.pathfmt.filename = name
 
@@ -916,8 +1093,8 @@ class ZipTest(BasePostprocessorTest):
 
         # write 3 files
         for i in range(3):
-            self.pathfmt.temppath = self.pathfmt.realdirectory + "file.ext"
-            self.pathfmt.filename = "file{}.ext".format(i)
+            self.pathfmt.temppath = f"{self.pathfmt.realdirectory}file.ext"
+            self.pathfmt.filename = f"file{i}.ext"
             self._trigger()
 
         # write the last file a second time (should be skipped)
@@ -927,8 +1104,8 @@ class ZipTest(BasePostprocessorTest):
         self._trigger(("finalize",))
 
         self.assertEqual(pp.zfile.write.call_count, 3)
-        for call in pp.zfile.write.call_args_list:
-            args, kwargs = call
+        for call_args in pp.zfile.write.call_args_list:
+            args, kwargs = call_args
             self.assertEqual(len(args), 2)
             self.assertEqual(len(kwargs), 0)
             self.assertEqual(args[0], self.pathfmt.temppath)
